@@ -1,7 +1,7 @@
 /* daemon.c -- Server daemon
  * Created: Fri Feb 28 18:17:56 1997 by faith@dict.org
- * Revised: Fri Dec 22 06:04:36 2000 by faith@dict.org
- * Copyright 1997, 1998, 1999, 2000 Rickard E. Faith (faith@dict.org)
+ * Revised: Tue Apr 23 09:14:45 2002 by faith@dict.org
+ * Copyright 1997, 1998, 1999, 2000, 2002 Rickard E. Faith (faith@dict.org)
  * 
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -17,22 +17,37 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: daemon.c,v 1.32 2000/12/22 14:15:25 faith Exp $
+ * $Id: daemon.c,v 1.60 2003/03/03 17:24:26 cheusov Exp $
  * 
  */
 
 #include "dictd.h"
 #include <ctype.h>
 #include <setjmp.h>
+
 #include "md5.h"
 #include "regex.h"
+#include "dictdplugin.h"
+#include "strategy.h"
+
+#ifndef HAVE_INET_ATON
+#define inet_aton(a,b) (b)->s_addr = inet_addr(a)
+#endif
+
+#ifndef HAVE_SNPRINTF
+extern int snprintf(char *str, size_t size, const char *format, ...);
+#endif
+
+#ifndef HAVE_VSNPRINTF
+extern int vsnprintf(char *str, size_t size, const char *format, va_list ap);
+#endif
+
 
 static int          _dict_defines, _dict_matches;
 static int          daemonS;
 static const char   *daemonHostname;
 static const char   *daemonIP;
 static int          daemonPort;
-static lst_Position databasePosition;
 static char         daemonStamp[256];
 static jmp_buf      env;
 static int          daemonMime;
@@ -52,22 +67,6 @@ static void daemon_status( const char *cmdline, int argc, char **argv );
 static void daemon_help( const char *cmdline, int argc, char **argv );
 static void daemon_quit( const char *cmdline, int argc, char **argv );
 
-static struct {
-   const char *name;
-   const char *description;
-   int        number;
-} strategyInfo[] = {
-   {"exact",     "Match words exactly",                        DICT_EXACT },
-   {"prefix",    "Match prefixes",                             DICT_PREFIX },
-   {"substring", "Match substring occurring anywhere in word", DICT_SUBSTRING},
-   {"suffix",    "Match suffixes",                             DICT_SUFFIX},
-   {"re",        "POSIX 1003.2 (modern) regular expressions",  DICT_RE },
-   {"regexp",    "Old (basic) regular expressions",            DICT_REGEXP },
-   {"soundex",   "Match using SOUNDEX algorithm",              DICT_SOUNDEX },
-   {"lev", "Match words within Levenshtein distance one", DICT_LEVENSHTEIN },
-};
-#define STRATEGIES (sizeof(strategyInfo)/sizeof(strategyInfo[0]))
-
 #define MAXARGCS 3
 static struct {
    int        argc;
@@ -98,21 +97,6 @@ static struct {
 };
 #define COMMANDS (sizeof(commandInfo)/sizeof(commandInfo[0]))
 
-
-static int lookup_strategy( const char *strategy )
-{
-   int i;
-
-   if (strategy[0] == '.' && strategy[1] == '\0')
-      return DICT_DEFAULT_STRATEGY;
-   
-   for (i = 0; i < STRATEGIES; i++) {
-      if (!strcasecmp(strategy, strategyInfo[i].name))
-         return strategyInfo[i].number;
-   }
-   return -1;
-}
-
 static void *(lookup_command)( int argc, char **argv )
 {
    int i, j;
@@ -132,51 +116,176 @@ static void *(lookup_command)( int argc, char **argv )
    return NULL;
 }
 
-static void reset_databases( void )
+static unsigned long daemon_compute_mask(int bits)
 {
-   databasePosition = lst_init_position( DictConfig->dbl );
+   unsigned long mask = 0xffffffff;
+
+   if (bits < 1) return 0;
+   if (bits < 32) mask -= (1 << (32-bits)) - 1;
+   return mask;
 }
 
-static dictDatabase *next_database( const char *name )
+static void daemon_log( int type, const char *format, ... )
 {
-   dictDatabase *db = NULL;
+   va_list ap;
+   char    buf[8*1024];
+   char    *buf2;
+   int     len;
+   char    *s, *d;
+   int     c;
+   char    marker = '?';
 
-   if (!name) return NULL;
-
-   if (*name == '*' || *name == '!') {
-      if (databasePosition) {
-	 do {
-	    db = lst_get_position( databasePosition );
-	    databasePosition = lst_next_position( databasePosition );
-	 } while (db && !db->available);
-      }
-      return db;
-   } else {
-      while (databasePosition) {
-         db = lst_get_position( databasePosition );
-	 databasePosition = lst_next_position( databasePosition );
-         if (db && !strcmp(db->databaseName,name)) {
-	    if (db->available) return db;
-	    else               return NULL;
-	 }
-      }
-      return NULL;
+   switch (type) {
+   case DICT_LOG_TERM:
+      if (!flg_test(LOG_STATS))    return; marker = 'I'; break;
+   case DICT_LOG_TRACE:
+      if (!flg_test(LOG_SERVER))   return; marker = 'I'; break;
+   case DICT_LOG_CLIENT:
+      if (!flg_test(LOG_CLIENT))   return; marker = 'C'; break;
+   case DICT_LOG_CONNECT:
+      if (!flg_test(LOG_CONNECT))  return; marker = 'K'; break;
+   case DICT_LOG_DEFINE:
+      if (!flg_test(LOG_FOUND))    return; marker = 'D'; break;
+   case DICT_LOG_MATCH:
+      if (!flg_test(LOG_FOUND))    return; marker = 'M'; break;
+   case DICT_LOG_NOMATCH:
+      if (!flg_test(LOG_NOTFOUND)) return; marker = 'N'; break;
+   case DICT_LOG_COMMAND:
+      if (!flg_test(LOG_COMMAND))  return; marker = 'T'; break;
+   case DICT_LOG_AUTH:
+      if (!flg_test(LOG_AUTH))     return; marker = 'A'; break;
    }
+
+   if (dbg_test(DBG_PORT))
+      snprintf(
+	 buf, sizeof (buf)/2,
+	 ":%c: %s:%d ", marker, daemonHostname, daemonPort );
+   else if (flg_test(LOG_HOST))
+      snprintf(
+	 buf, sizeof (buf)/2,
+	 ":%c: %s ", marker, daemonHostname );
+   else
+      snprintf(
+	 buf, sizeof (buf)/2,
+	 ":%c: ", marker );
+
+   len = strlen( buf );
+
+   va_start( ap, format );
+   vsnprintf( buf+len, sizeof (buf)/2-len, format, ap );
+   va_end( ap );
+   len = strlen( buf );
+
+   if (len >= sizeof (buf)/2) {
+      log_info( ":E: buffer overflow (%d)\n", len );
+      buf[2048] = '\0';
+      len = strlen(buf);
+   }
+
+   buf2 = alloca( 3*(len+3) );
+
+   for (s = buf, d = buf2; *s; s++) {
+      c = (unsigned char)*s;
+      if (c == '\t')      *d++ = ' ';
+      else if (c == '\n') *d++ = c;
+      else {
+	 if (c < 32)        { *d++ = '^'; *d++ = c + 64;           }
+	 else if (c == 127) { *d++ = '^'; *d++ = '?';              }
+	 else                 *d++ = c;
+      }
+   }
+   *d = '\0';
+   log_info( "%s", buf2 );
+
+   if (d != buf2) d[-1] = '\0';	/* kill newline */
+   dict_setproctitle( "dictd %s", buf2 );
 }
 
-static int count_databases( void )
+static int daemon_check_mask(const char *spec, const char *ip)
 {
-   int count = 0;
+   char           *tmp = alloca(strlen(spec) + 1);
+   char           *pt;
+   char           tstring[64], mstring[64];
+   struct in_addr target, mask;
+   int            bits;
+   unsigned long  bitmask;
    
-   reset_databases();
-   while (next_database("*")) ++count;
-   return count;
+   strcpy(tmp, spec);
+   if (!(pt = strchr(tmp, '/'))) {
+      log_info( ":E: No / in %s, denying access to %s\n", spec, ip);
+      return DICT_DENY;
+   }
+   *pt++ = '\0';
+   if (!*pt) {
+      log_info( ":E: No bit count after / in %s, denying access to %s\n",
+                spec, ip);
+      return DICT_DENY;
+   }
+   
+   inet_aton(ip, &target);
+   inet_aton(tmp, &mask);
+   bits = strtol(pt, NULL, 10);
+   strcpy(tstring, inet_ntoa(target));
+   strcpy(mstring, inet_ntoa(mask));
+   if (bits < 0 || bits > 32) {
+      log_info( ":E: Bit count (%d) out of range, denying access to %s\n",
+                bits, ip);
+      return DICT_DENY;
+   }
+   
+   bitmask = daemon_compute_mask(bits);
+   if ((ntohl(target.s_addr) & bitmask) == (ntohl(mask.s_addr) & bitmask)) {
+      PRINTF(DBG_AUTH, ("%s matches %s/%d\n", tstring, mstring, bits));
+      return DICT_MATCH;
+   }
+   PRINTF(DBG_AUTH, ("%s does NOT match %s/%d\n", tstring, mstring, bits));
+   return DICT_NOMATCH;
+   
+   return 0;
 }
 
-static int daemon_check_list( const char *user, lst_List acl )
+static int daemon_check_range(const char *spec, const char *ip)
+{
+   char           *tmp = alloca(strlen(spec) + 1);
+   char           *pt;
+   char           tstring[64], minstring[64], maxstring[64];
+   struct in_addr target, min, max;
+   
+   strcpy(tmp, spec);
+   if (!(pt = strchr(tmp, ':'))) {
+      log_info( ":E: No : in range %s, denying access to %s\n", spec, ip);
+      return DICT_DENY;
+   }
+   *pt++ = '\0';
+   if (strchr(pt, ':')) {
+      log_info( ":E: More than one : in range %s, denying access to %s\n",
+                spec, ip);
+      return DICT_DENY;
+   }
+   if (!*pt) {
+      log_info( ":E: Misformed range %s, denying access to %s\n", spec, ip);
+      return DICT_DENY;
+   }
+   
+   inet_aton(ip, &target);
+   inet_aton(tmp, &min);
+   inet_aton(pt, &max);
+   strcpy(tstring, inet_ntoa(target));
+   strcpy(minstring, inet_ntoa(min));
+   strcpy(maxstring, inet_ntoa(max));
+   if (ntohl(target.s_addr) >= ntohl(min.s_addr)
+       && ntohl(target.s_addr) <= ntohl(max.s_addr)) {
+      PRINTF(DBG_AUTH,("%s in range from %s to %s\n",
+                       tstring, minstring, maxstring));
+      return DICT_MATCH;
+   }
+   PRINTF(DBG_AUTH,("%s NOT in range from %s to %s\n",
+                    tstring, minstring, maxstring));
+   return DICT_NOMATCH;
+}
+
+static int daemon_check_wildcard(const char *spec, const char *ip)
 {
-   lst_Position p;
-   dictAccess   *a;
    char         regbuf[256];
    char         erbuf[100];
    int          err;
@@ -184,6 +293,39 @@ static int daemon_check_list( const char *user, lst_List acl )
    char         *d;
    regex_t      re;
 
+   for (d = regbuf, s = spec; s && *s; ++s) {
+      switch (*s) {
+      case '*': *d++ = '.';  *d++ = '*'; break;
+      case '.': *d++ = '\\'; *d++ = '.'; break;
+      case '?': *d++ = '.';              break;
+      default:  *d++ = *s;               break;
+      }
+   }
+   *d = '\0';
+   if ((err = regcomp(&re, regbuf, REG_ICASE|REG_NOSUB))) {
+      regerror(err, &re, erbuf, sizeof(erbuf));
+      log_info( ":E: regcomp(%s): %s\n", regbuf, erbuf );
+      return DICT_DENY;	/* Err on the side of safety */
+   }
+   if (!regexec(&re, daemonHostname, 0, NULL, 0)
+       || !regexec(&re, daemonIP, 0, NULL, 0)) {
+      PRINTF(DBG_AUTH,
+             ("Match %s with %s/%s\n", spec, daemonHostname, daemonIP));
+      regfree(&re);
+      return DICT_MATCH;
+   }
+   regfree(&re);
+   PRINTF(DBG_AUTH,
+          ("No match (%s with %s/%s)\n", spec, daemonHostname, daemonIP));
+   return DICT_NOMATCH;
+}
+
+static int daemon_check_list( const char *user, lst_List acl )
+{
+   lst_Position p;
+   dictAccess   *a;
+   int          retcode;
+
    if (!acl) return DICT_ALLOW;
    for (p = lst_init_position(acl); p; p = lst_next_position(p)) {
       a = lst_get_position(p);
@@ -191,35 +333,25 @@ static int daemon_check_list( const char *user, lst_List acl )
       case DICT_DENY:
       case DICT_ALLOW:
       case DICT_AUTHONLY:
-	 for (d = regbuf, s = a->spec; s && *s; ++s) {
-	    switch (*s) {
-	    case '*': *d++ = '.';  *d++ = '*'; break;
-	    case '.': *d++ = '\\'; *d++ = '.'; break;
-	    case '?': *d++ = '.';              break;
-	    default:  *d++ = *s;               break;
-	    }
-	 }
-	 *d = '\0';
-	 if ((err = regcomp(&re, regbuf, REG_ICASE|REG_NOSUB))) {
-	    regerror(err, &re, erbuf, sizeof(erbuf));
-	    log_info( ":E: regcomp(%s): %s\n", regbuf, erbuf );
-	    return DICT_DENY;	/* Err on the side of safety */
-	 }
-	 if (!regexec(&re, daemonHostname, 0, NULL, 0)
-	     || !regexec(&re, daemonIP, 0, NULL, 0)) {
-	    PRINTF(DBG_AUTH,
-		   ("Match %s with %s/%s\n",
-		    a->spec,daemonHostname,daemonIP));
-	    regfree(&re);
-	    return a->type;
-	 }
-	 regfree(&re);
-	 PRINTF(DBG_AUTH,
-		("No match (%s with %s/%s)\n",
-		 a->spec,daemonHostname,daemonIP));
+         if (strchr(a->spec, '/'))
+            retcode = daemon_check_mask(a->spec, daemonIP);
+         else if  (strchr(a->spec, ':'))
+            retcode = daemon_check_range(a->spec, daemonIP);
+         else
+            retcode = daemon_check_wildcard(a->spec, daemonIP);
+         switch (retcode) {
+         case DICT_DENY:
+            return DICT_DENY;
+         case DICT_MATCH:
+            if (a->type == DICT_DENY) {
+                daemon_log( DICT_LOG_AUTH, "spec %s denies %s/%s\n",
+                            a->spec, daemonHostname, daemonIP);
+            }
+            return a->type;
+         }
 	 break;
       case DICT_USER:
-	 if (!strcmp(user,a->spec)) return DICT_ALLOW;
+	 if (user && !strcmp(user,a->spec)) return DICT_ALLOW;
       case DICT_GROUP:		/* Groups are not yet implemented. */
 	 break;
       }
@@ -252,73 +384,6 @@ static int daemon_check_auth( const char *user )
    return 0;
 }
 
-static void daemon_log( int type, const char *format, ... )
-{
-   va_list ap;
-   char    buf[8*1024];
-   char    *buf2;
-   int     len;
-   char    *s, *d;
-   int     c;
-   char    marker = '?';
-
-   switch (type) {
-   case DICT_LOG_TERM:
-      if (!flg_test(LOG_STATS))    return; marker = 'I'; break;
-   case DICT_LOG_TRACE:
-      if (!flg_test(LOG_SERVER))   return; marker = 'I'; break;
-   case DICT_LOG_CLIENT:
-      if (!flg_test(LOG_CLIENT))   return; marker = 'C'; break;
-   case DICT_LOG_DEFINE:
-      if (!flg_test(LOG_FOUND))    return; marker = 'D'; break;
-   case DICT_LOG_MATCH:
-      if (!flg_test(LOG_FOUND))    return; marker = 'M'; break;
-   case DICT_LOG_NOMATCH:
-      if (!flg_test(LOG_NOTFOUND)) return; marker = 'N'; break;
-   case DICT_LOG_COMMAND:
-      if (!flg_test(LOG_COMMAND))  return; marker = 'T'; break;
-   }
-
-   if (dbg_test(DBG_PORT))
-      sprintf( buf, ":%c: %s:%d ", marker, daemonHostname, daemonPort );
-   else if (flg_test(LOG_HOST))
-      sprintf( buf, ":%c: %s ", marker, daemonHostname );
-   else
-      sprintf( buf, ":%c: ", marker );
-      
-   len = strlen( buf );
-   
-   va_start( ap, format );
-   vsprintf( buf+len, format, ap );
-   va_end( ap );
-   len = strlen( buf );
-
-   if (len > 2048) {
-      log_info( ":E: buffer overflow (%d)\n", len );
-      buf[2048] = '\0';
-      len = strlen(buf);
-   }
-
-   buf2 = alloca( 3*(len+3) );
-
-   for (s = buf, d = buf2; *s; s++) {
-      c = (unsigned char)*s;
-      if (c == '\t')      *d++ = ' ';
-      else if (c == '\n') *d++ = c;
-      else {
-	 if (c > 128)       { *d++ = 'M'; *d++ = '-';    c -= 128; }
-	 if (c < 32)        { *d++ = '^'; *d++ = c + 64;           }
-	 else if (c == 127) { *d++ = '^'; *d++ = '?';              }
-	 else                 *d++ = c;
-      }
-   }
-   *d = '\0';
-   log_info( "%s", buf2 );
-
-   if (d != buf2) d[-1] = '\0';	/* kill newline */
-   dict_setproctitle( "dictd %s", buf2 );
-}
-
 void daemon_terminate( int sig, const char *name )
 {
    alarm(0);
@@ -345,7 +410,7 @@ void daemon_terminate( int sig, const char *name )
                   dict_format_time( tim_get_user( "t" ) ),
                   dict_format_time( tim_get_system( "t" ) ) );
    }
-   
+   log_close();
    longjmp(env,1);
    if (sig) exit(sig+128);
    exit(0);
@@ -412,7 +477,7 @@ static void daemon_printf( const char *format, ... )
    int     len;
 
    va_start( ap, format );
-   vsprintf( buf, format, ap );
+   vsnprintf( buf, sizeof (buf), format, ap );
    va_end( ap );
    if ((len = strlen( buf )) >= BUFFERSIZE) {
       log_info( ":E: buffer overflow: %d\n", len );
@@ -480,12 +545,22 @@ static int daemon_count_defs( lst_List list )
    dictWord      *dw;
    unsigned long previousStart = 0;
    unsigned long previousEnd   = 0;
+   const char   *previousDef   = NULL;
    int           count         = 0;
 
    LST_ITERATE(list,p,dw) {
-      if (previousStart == dw->start && previousEnd == dw->end) continue;
+      if (
+	 previousStart == dw->start &&
+	 previousEnd   == dw->end &&
+	 previousDef   == dw->def)
+      {
+	 continue;
+      }
+
       previousStart = dw->start;
       previousEnd   = dw->end;
+      previousDef   = dw->def;
+
       ++count;
    }
    return count;
@@ -496,24 +571,39 @@ static void daemon_dump_defs( lst_List list )
    lst_Position  p;
    char          *buf;
    dictWord      *dw;
-   dictDatabase  *db;
+   const dictDatabase  *db;
    unsigned long previousStart = 0;
    unsigned long previousEnd   = 0;
+   const char *  previousDef   = NULL;
 
    LST_ITERATE(list,p,dw) {
       db = dw->database;
 
-      if (previousStart == dw->start && previousEnd == dw->end) continue;
+      if (
+	 previousStart == dw->start &&
+	 previousEnd   == dw->end &&
+	 previousDef   == dw->def) 
+      {
+	 continue;
+      }
+
       previousStart = dw->start;
       previousEnd   = dw->end;
-      buf = dict_data_read( db->data, dw->start, dw->end,
-			    db->prefilter, db->postfilter );
+      previousDef   = dw->def;
+
+      buf = dict_data_obtain ( db, dw );
+
+      if (dw -> database_visible){
+	 db = dw -> database_visible;
+      }
+
+      daemon_printf (
+	  "%d \"%s\" %s \"%s\"\n",
+	  CODE_DEFINITION_FOLLOWS,
+	  dw->word,
+	  db->invisible ? "*" : db->databaseName,
+	  db->invisible ? ""  : db->databaseShort);
 
-      daemon_printf( "%d \"%s\" %s \"%s\"\n",
-		     CODE_DEFINITION_FOLLOWS,
-		     dw->word,
-		     db->databaseName,
-		     db->databaseShort );
       daemon_mime();
       daemon_text(buf);
       xfree( buf );
@@ -524,15 +614,17 @@ static int daemon_count_matches( lst_List list )
 {
    lst_Position p;
    dictWord     *dw;
-   const char   *prevword = NULL;
-   dictDatabase *prevdb   = NULL;
-   int          count     = 0;
-   
+   const char   *prevword     = NULL;
+   const dictDatabase *prevdb = NULL;
+   int           count        = 0;
+
    LST_ITERATE(list,p,dw) {
       if (prevdb == dw->database && prevword && !strcmp(prevword,dw->word))
-	  continue;
+	 continue;
+
       prevword = dw->word;
       prevdb   = dw->database;
+
       ++count;
    }
    return count;
@@ -542,16 +634,28 @@ static void daemon_dump_matches( lst_List list )
 {
    lst_Position p;
    dictWord     *dw;
-   const char   *prevword = NULL;
-   dictDatabase *prevdb   = NULL;
-   
+   const char   *prevword     = NULL;
+   const dictDatabase *prevdb = NULL;
+   const dictDatabase *db     = NULL;
+
    daemon_mime();
    LST_ITERATE(list,p,dw) {
+      db = dw -> database;
+
       if (prevdb == dw->database && prevword && !strcmp(prevword,dw->word))
 	  continue;
+
       prevword = dw->word;
       prevdb   = dw->database;
-      daemon_printf( "%s \"%s\"\n", dw->database->databaseName, dw->word );
+
+      if (dw -> database_visible){
+	 db = dw -> database_visible;
+      }
+
+      daemon_printf (
+	  "%s \"%s\"\n",
+	  db -> invisible ? "*" : db -> databaseName,
+	  dw -> word );
    }
    daemon_printf( ".\n" );
 }
@@ -562,7 +666,7 @@ static void daemon_banner( void )
 
    time(&t);
 
-   sprintf( daemonStamp, "<%d.%d.%lu@%s>", 
+   snprintf( daemonStamp, sizeof (daemonStamp), "<%d.%d.%lu@%s>", 
 	    _dict_forks,
 	    getpid(),
 	    (long unsigned)t,
@@ -577,12 +681,11 @@ static void daemon_banner( void )
 static void daemon_define( const char *cmdline, int argc, char **argv )
 {
    lst_List       list = lst_create();
-   dictDatabase   *db;
    int            matches = 0;
    char           *word;
    const char     *databaseName;
-   int            none = 1;
    int            extension = (argv[0][0] == 'd' && argv[0][1] == '\0');
+   int            db_found = 0;
 
    if (extension) {
       switch (argc) {
@@ -602,18 +705,14 @@ static void daemon_define( const char *cmdline, int argc, char **argv )
       return;
    }
 
-   reset_databases();
-   while ((db = next_database(databaseName))) {
-      none = 0;
-      matches += dict_search_database( list, word, db, DICT_EXACT );
-      if (matches > 0 && *databaseName != '*')               break; 
-      else if (*databaseName == '*' || *databaseName == '!') continue;
-      goto nomatch;
-   }
+   matches = abs(dict_search_databases (
+      list, NULL,
+      databaseName, word, DICT_EXACT,
+      &db_found));
 
-   if (matches) {
+   if (db_found && matches > 0) {
       int actual_matches = daemon_count_defs( list );
-      
+
       _dict_defines += actual_matches;
       daemon_log( DICT_LOG_DEFINE,
 		  "%s \"%s\" %d\n", databaseName, word, actual_matches);
@@ -625,14 +724,14 @@ static void daemon_define( const char *cmdline, int argc, char **argv )
       dict_destroy_list( list );
       return;
    }
-   
-   if (none) {
+
+   if (!db_found) {
+      dict_destroy_list( list );
       daemon_printf( "%d invalid database, use SHOW DB for list\n",
 		     CODE_INVALID_DB );
       return;
    }
-   
- nomatch:
+
    dict_destroy_list( list );
    daemon_log( DICT_LOG_NOMATCH,
 	       "%s exact \"%s\"\n", databaseName, word );
@@ -642,14 +741,13 @@ static void daemon_define( const char *cmdline, int argc, char **argv )
 static void daemon_match( const char *cmdline, int argc, char **argv )
 {
    lst_List       list = lst_create();
-   dictDatabase   *db;
    int            matches = 0;
    char           *word;
    const char     *databaseName;
    const char     *strategy;
    int            strategyNumber;
-   int            none = 1;
    int            extension = (argv[0][0] == 'm' && argv[0][1] == '\0');
+   int            db_found = 0;
 
    if (extension) {
       switch (argc) {
@@ -657,9 +755,9 @@ static void daemon_match( const char *cmdline, int argc, char **argv )
       case 3:databaseName = "*";     strategy = argv[1]; word = argv[2]; break;
       case 4:databaseName = argv[1]; strategy = argv[2]; word = argv[3]; break;
       default:
-      daemon_printf( "%d syntax error, illegal parameters\n",
-		     CODE_ILLEGAL_PARAM );
-      return;
+	 daemon_printf( "%d syntax error, illegal parameters\n",
+			CODE_ILLEGAL_PARAM );
+	 return;
       }
    } else if (argc == 4) {
       databaseName = argv[1];
@@ -677,16 +775,12 @@ static void daemon_match( const char *cmdline, int argc, char **argv )
       return;
    }
 
-   reset_databases();
-   while ((db = next_database(databaseName))) {
-      none = 0;
-      matches += dict_search_database( list, word, db, strategyNumber );
-      if (matches > 0 && *databaseName != '*')               break;
-      else if (*databaseName == '*' || *databaseName == '!') continue;
-      goto nomatch;
-   }
+   matches = abs(dict_search_databases (
+      list, NULL,
+      databaseName, word, strategyNumber | DICT_MATCH_MASK,
+      &db_found));
 
-   if (matches) {
+   if (db_found && matches > 0) {
       int actual_matches = daemon_count_matches( list );
       
       _dict_matches += actual_matches;
@@ -700,24 +794,270 @@ static void daemon_match( const char *cmdline, int argc, char **argv )
       dict_destroy_list( list );
       return;
    }
-   
-   if (none) {
+
+   if (!db_found) {
+      dict_destroy_list( list );
       daemon_printf( "%d invalid database, use SHOW DB for list\n",
 		     CODE_INVALID_DB );
       return;
    }
 
- nomatch:
+   dict_destroy_list( list );
    daemon_log( DICT_LOG_NOMATCH,
 	       "%s %s \"%s\"\n", databaseName, strategy, word );
    daemon_ok( CODE_NO_MATCH, "no match", "c" );
 }
 
+static lst_Position first_database_pos (void)
+{
+   return lst_init_position (DictConfig->dbl);
+}
+
+static dictDatabase *next_database (
+   lst_Position *databasePosition,
+   const char *name)
+{
+   dictDatabase *db = NULL;
+
+   assert (databasePosition);
+
+   if (!name)
+      return NULL;
+
+   if (*name == '*' || *name == '!') {
+      if (*databasePosition) {
+	 do {
+	    db = lst_get_position( *databasePosition );
+	    *databasePosition = lst_next_position( *databasePosition );
+	 } while (db && !db->available);
+      }
+      return db;
+   } else {
+      while (*databasePosition) {
+         db = lst_get_position( *databasePosition );
+	 *databasePosition = lst_next_position( *databasePosition );
+         if (db && !strcmp(db -> databaseName,name) && !db -> invisible) {
+	    if (db->available)
+	       return db;
+	    else
+               return NULL;
+	 }
+      }
+      return NULL;
+   }
+}
+
+static int count_databases( void )
+{
+   int count = 0;
+   const dictDatabase *db;
+
+   lst_Position databasePosition = first_database_pos ();
+
+   while (NULL != (db = next_database (&databasePosition, "*"))){
+      if (!db->invisible){
+	 ++count;
+      }
+   }
+
+   return count;
+}
+
+static void destroy_word_list (lst_List l)
+{
+   char *word;
+
+   while (lst_length (l)){
+      word = lst_pop (l);
+      if (word)
+	 xfree (word);
+   }
+   lst_destroy (l);
+}
+
+/*
+  Replaces invisible databases with db argument.
+ */
+static void replace_invisible_databases (
+   lst_Position *pos,
+   const dictDatabase *db)
+{
+   dictWord *dw;
+
+   while (pos){
+      dw = (dictWord *) lst_get_position (pos);
+
+      if (dw -> database && dw -> database -> invisible){
+	 dw -> database_visible = db;
+      }
+
+      pos = lst_next_position (pos);
+   }
+}
+
+/*
+  Search for all words in word_list in the database db
+ */
+static int dict_search_words (
+   lst_List *l,
+   lst_List word_list,
+   const dictDatabase *db,
+   int strategy,
+   int *error, int *result,
+   const dictPluginData **extra_result, int *extra_result_size)
+{
+   lst_Position word_list_pos;
+   int mc = 0;
+   int matches_count = 0;
+   const char *word;
+
+   int old_count = 0;
+   int db_found  = 0;
+
+   word_list_pos = lst_init_position (word_list);
+   while (word_list_pos){
+      word = lst_get_position (word_list_pos);
+
+      if (word){
+	 if (db -> virtual_db_list){
+	    assert (lst_init_position (db -> virtual_db_list));
+
+	    old_count = lst_length (l);
+
+	    matches_count = dict_search_databases (
+	       l, lst_init_position (db -> virtual_db_list),
+	       "*", word, strategy, &db_found);
+
+	    assert (db_found);
+
+	    if (matches_count > 0){
+	       replace_invisible_databases (
+		  lst_nth_position (l, old_count + 1),
+		  db);
+	    }
+	 }else{
+	    matches_count = dict_search (
+	       l, word, db, strategy,
+	       result, extra_result, extra_result_size);
+
+	    if (*result == DICT_PLUGIN_RESULT_PREPROCESS){
+	       assert (matches_count > 0);
+
+	       xfree (lst_get_position (word_list_pos));
+	       lst_set_position (word_list_pos, NULL);
+	    }
+	 }
+
+	 if (matches_count < 0){
+	    *error = 1;
+	    matches_count = abs (matches_count);
+	    mc += matches_count;
+	    break;
+	 }
+
+	 mc += matches_count;
+      }
+
+      word_list_pos = lst_next_position (word_list_pos);
+   }
+
+   return mc;
+}
+
+int dict_search_databases (
+   lst_List *l,
+   lst_Position databasePosition, /* NULL for global database list */
+   const char *databaseName, const char *word, int strategy,
+   int *db_found)
+{
+   int matches       = -1;
+   int matches_count = 0;
+   int error         = 0;
+
+
+   const dictDatabase *db;
+   dictWord *dw;
+   char *p;
+
+   lst_List preprocessed_words;
+   int i;
+
+   int                   result;
+   const dictPluginData *extra_result;
+   int                   extra_result_size;
+
+   *db_found = 0;
+
+   if (!databasePosition)
+      databasePosition = first_database_pos ();
+
+   preprocessed_words = lst_create ();
+   lst_append (preprocessed_words, xstrdup(word));
+
+   while (!error && (db = next_database (&databasePosition, databaseName))) {
+      if (db -> exit_db)
+	 /* dictionary_exit */
+	 break;
+
+      *db_found = 1;
+
+      result = DICT_PLUGIN_RESULT_NOTFOUND;
+
+      matches_count = dict_search_words (
+	 l,
+	 preprocessed_words, db, strategy,
+	 &error, &result, &extra_result, &extra_result_size);
+
+      if (matches < 0)
+	 matches = 0;
+
+      if (result == DICT_PLUGIN_RESULT_PREPROCESS){
+	 for (i=0; i < matches_count; ++i){
+	    dw = lst_pop (l);
+	    switch (dw -> def_size){
+	    case -1:
+	       p = xstrdup (dw -> word);
+	       lst_append (preprocessed_words, p);
+	       break;
+	    case 0:
+	       break;
+	    default:
+	       p = xmalloc (1 + dw -> def_size);
+	       memcpy (p, dw -> def, dw -> def_size);
+	       p [dw -> def_size] = 0;
+
+	       lst_append (preprocessed_words, p);
+	    }
+
+	    dict_destroy_datum (dw);
+	 }
+      }else{
+	 matches += matches_count;
+
+	 if (result == DICT_PLUGIN_RESULT_EXIT)
+	    break;
+
+	 if (*databaseName == '*')
+	    continue;
+	 else if (!matches && *databaseName == '!')
+	    continue;
+
+	 break;
+      }
+   }
+
+   destroy_word_list (preprocessed_words);
+
+   return error ? -matches : matches;
+}
+
 static void daemon_show_db( const char *cmdline, int argc, char **argv )
 {
    int          count;
-   dictDatabase *db;
+   const dictDatabase *db;
    
+   lst_Position databasePosition;
+
    if (argc != 2) {
       daemon_printf( "%d syntax error, illegal parameters\n",
 		     CODE_ILLEGAL_PARAM );
@@ -729,11 +1069,15 @@ static void daemon_show_db( const char *cmdline, int argc, char **argv )
    } else {
       daemon_printf( "%d %d databases present\n",
 		     CODE_DATABASE_LIST, count );
-      reset_databases();
+
+      databasePosition = first_database_pos ();
+
       daemon_mime();
-      while ((db = next_database("*"))) {
-	 daemon_printf( "%s \"%s\"\n",
-			db->databaseName, db->databaseShort );
+      while ((db = next_database(&databasePosition, "*"))) {
+	 if (!db->invisible){
+	    daemon_printf( "%s \"%s\"\n",
+			   db->databaseName, db->databaseShort );
+	 }
       }
       daemon_printf( ".\n" );
       daemon_ok( CODE_OK, "ok", NULL );
@@ -743,25 +1087,30 @@ static void daemon_show_db( const char *cmdline, int argc, char **argv )
 static void daemon_show_strat( const char *cmdline, int argc, char **argv )
 {
    int i;
-   
+
+   int strat_count        = get_strategy_count ();
+   dictStrategy ** strats = get_strategies ();
+
    if (argc != 2) {
       daemon_printf( "%d syntax error, illegal parameters\n",
 		     CODE_ILLEGAL_PARAM );
       return;
    }
 
-   if (!STRATEGIES) {
-      daemon_printf( "%d no strategies available\n", CODE_NO_STRATEGIES );
-   } else {
+   if (strat_count){
       daemon_printf( "%d %d databases present\n",
-		     CODE_STRATEGY_LIST, STRATEGIES );
+		     CODE_STRATEGY_LIST, strat_count );
       daemon_mime();
-      for (i = 0; i < STRATEGIES; i++) {
+
+      for (i = 0; i < strat_count; i++) {
 	 daemon_printf( "%s \"%s\"\n",
-			strategyInfo[i].name, strategyInfo[i].description );
+			strats [i] -> name, strats [i] -> description );
       }
+
       daemon_printf( ".\n" );
       daemon_ok( CODE_OK, "ok", NULL );
+   }else{
+      daemon_printf( "%d no strategies available\n", CODE_NO_STRATEGIES );
    }
 }
 
@@ -769,9 +1118,12 @@ static void daemon_show_info( const char *cmdline, int argc, char **argv )
 {
    char         *buf;
    dictWord     *dw;
-   dictDatabase *db;
+   const dictDatabase *db;
    lst_List     list;
-   
+   const char  *info_entry_name = DICT_INFO_ENTRY_NAME;
+
+   lst_Position databasePosition = first_database_pos ();
+
    if (argc != 3) {
       daemon_printf( "%d syntax error, illegal parameters\n",
 		     CODE_ILLEGAL_PARAM );
@@ -785,17 +1137,33 @@ static void daemon_show_info( const char *cmdline, int argc, char **argv )
    }
 
    list = lst_create();
-   reset_databases();
-   while ((db = next_database( argv[2] ))) {
-      if (dict_search_database( list,
-				db->databaseInfoPointer ?
-				db->databaseInfoPointer :
-				DICT_INFO_ENTRY_NAME,
-				db,
-				DICT_EXACT )) {
+   while ((db = next_database(&databasePosition, argv[2] ))) {
+      if (
+	 db -> databaseInfo &&
+	 (db -> virtual_db || db -> databaseInfo [0] != '@'))
+      {
+	 daemon_printf( "%d information for %s\n",
+			CODE_DATABASE_INFO, argv[2] );
+	 daemon_mime();
+	 daemon_text(db -> databaseInfo);
+	 daemon_text("\n");
+	 daemon_ok( CODE_OK, "ok", NULL );
+	 return;
+      }
+
+      if (db -> databaseInfo && db -> databaseInfo [0] == '@')
+	 info_entry_name = db -> databaseInfo + 1;
+
+      if (dict_search (
+	 list,
+	 info_entry_name,
+	 db,
+	 DICT_EXACT,
+	 NULL, NULL, NULL))
+      {
 	 dw = lst_nth_get( list, 1 );
-	 buf = dict_data_read( db->data, dw->start, dw->end,
-			       db->prefilter, db->postfilter );
+	 buf = dict_data_obtain( db, dw );
+
 	 dict_destroy_list( list );
 	 daemon_printf( "%d information for %s\n",
 			CODE_DATABASE_INFO, argv[2] );
@@ -823,9 +1191,11 @@ static void daemon_show_server( const char *cmdline, int argc, char **argv )
 {
    FILE          *str;
    char          buffer[1024];
-   dictDatabase  *db;
+   const dictDatabase  *db;
    double        uptime;
    
+   lst_Position databasePosition = first_database_pos ();
+
    tim_stop("dictd");
    uptime = tim_get_real("dictd");
    
@@ -842,25 +1212,52 @@ static void daemon_show_server( const char *cmdline, int argc, char **argv )
    if (count_databases()) {
       daemon_printf( "\nDatabase      Headwords         Index"
 		     "          Data  Uncompressed\n" );
-      reset_databases();
-      while ((db = next_database("*"))) {
-	 daemon_printf( "%-12.12s %10lu %10lu %cB %10lu %cB %10lu %cB\n",
-			db->databaseName,
-			db->index->headwords,
-			
-			db->index->size/1024 > 10240 ?
-			db->index->size/1024/1024 : db->index->size/1024,
-			db->index->size/1024 > 10240 ? 'M' : 'k',
+      while ((db = next_database (&databasePosition, "*"))) {
+	 int headwords       = db->index ? db->index->headwords : 0;
 
-			db->data->size/1024 > 10240 ?
-			db->data->size/1024/1024 : db->data->size/1024,
-			db->data->size/1024 > 10240 ? 'M' : 'k',
+	 int index_size      = 0;
+	 char index_size_uom = 'k';
 
-			db->data->length/1024 > 10240 ?
-			db->data->length/1024/1024 : db->data->length/1024,
-			db->data->length/1024 > 10240 ? 'M' : 'k' );
+	 int data_size       = 0;
+	 char data_size_uom  = 'k';
+	 int data_length     = 0;
+	 char data_length_uom= 'k';
+
+	 if (db -> invisible)
+	    continue;
+
+	 if (db->index){
+	    index_size = db->index->size/1024 > 10240 ?
+	       db->index->size/1024/1024 : db->index->size/1024;
+	    index_size_uom = db->index->size/1024 > 10240 ? 'M' : 'k';
+	 }
+
+	 if (db->data){
+	    data_size = db->data->size/1024 > 10240 ?
+	       db->data->size/1024/1024 : db->data->size/1024;
+	    data_size_uom = db->data->size/1024 > 10240 ? 'M' : 'k';
+
+	    data_length = db->data->length/1024 > 10240 ?
+	       db->data->length/1024/1024 : db->data->length/1024;
+	    data_length_uom = db->data->length/1024 > 10240 ? 'M' : 'k';
+	 }
+
+	 daemon_printf(
+	    "%-12.12s %10lu %10lu %cB %10lu %cB %10lu %cB\n",
+	    db->databaseName,
+	    headwords,
+
+	    index_size,
+	    index_size_uom,
+
+	    data_size,
+	    data_size_uom,
+
+	    data_size,
+	    data_size_uom);
       }
    }
+
    if (DictConfig->site && (str = fopen( DictConfig->site, "r" ))) {
       daemon_printf( "\nSite-specific information for %s:\n\n",
 		     net_hostname() );
@@ -909,28 +1306,36 @@ static void daemon_auth( const char *cmdline, int argc, char **argv )
    unsigned char     digest[16];
    char              hex[33];
    int               i;
-   
+   int               buf_size;
+
    if (argc != 3)
       daemon_printf( "%d syntax error, illegal parameters\n",
 		     CODE_ILLEGAL_PARAM );
    if (!h || !(secret = hsh_retrieve(h, argv[1]))) {
+      daemon_log( DICT_LOG_AUTH, "%s@%s/%s denied: no invalid username\n",
+                  argv[1], daemonHostname, daemonIP );
       daemon_printf( "%d auth denied\n", CODE_AUTH_DENIED );
       return;
    }
-   
-   buf = alloca(strlen(daemonStamp) + strlen(secret) + 10);
-   sprintf( buf, "%s%s", daemonStamp, secret );
+
+   buf_size = strlen(daemonStamp) + strlen(secret) + 10;
+   buf = alloca(buf_size);
+   snprintf( buf, buf_size, "%s%s", daemonStamp, secret );
 
    MD5Init(&ctx);
    MD5Update(&ctx, buf, strlen(buf));
    MD5Final(digest, &ctx);
 
-   for (i = 0; i < 16; i++) sprintf( hex+2*i, "%02x", digest[i] );
+   for (i = 0; i < 16; i++)
+      snprintf( hex+2*i, 2, "%02x", digest[i] );
+
    hex[32] = '\0';
 
    PRINTF(DBG_AUTH,("Got %s expected %s\n", argv[2], hex ));
 
    if (strcmp(hex,argv[2])) {
+      daemon_log( DICT_LOG_AUTH, "%s@%s/%s denied: hash mismatch\n",
+                  argv[1], daemonHostname, daemonIP );
       daemon_printf( "%d auth denied\n", CODE_AUTH_DENIED );
    } else {
       daemon_printf( "%d authenticated\n", CODE_AUTH_OK );
@@ -1013,6 +1418,8 @@ int dict_daemon( int s, struct sockaddr_in *csin, char ***argv0, int delay,
 
    tim_start( "t" );
    daemon_log( DICT_LOG_TRACE, "connected\n" );
+   daemon_log( DICT_LOG_CONNECT, "%s/%s connected on port %d\n",
+               daemonHostname, daemonIP, daemonPort );
    dict_setproctitle( "dictd: %s connected", daemonHostname );
 
    if (error) {
@@ -1022,6 +1429,8 @@ int dict_daemon( int s, struct sockaddr_in *csin, char ***argv0, int delay,
    }
 
    if (daemon_check_auth( NULL )) {
+      daemon_log( DICT_LOG_AUTH, "%s/%s denied: ip/hostname rules\n",
+                  daemonHostname, daemonIP );
       daemon_printf( "%d access denied\n", CODE_ACCESS_DENIED );
       daemon_terminate( 0, "access denied" );
    }
