@@ -1,7 +1,7 @@
 /* data.c -- 
  * Created: Tue Jul 16 12:45:41 1996 by faith@dict.org
- * Revised: Fri Dec 22 06:04:44 2000 by faith@dict.org
- * Copyright 1996, 1997, 1998, 2000 Rickard E. Faith (faith@dict.org)
+ * Revised: Sat Mar 30 10:46:06 2002 by faith@dict.org
+ * Copyright 1996, 1997, 1998, 2000, 2002 Rickard E. Faith (faith@dict.org)
  * 
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -17,18 +17,31 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: data.c,v 1.9 2000/12/22 14:15:25 faith Exp $
+ * $Id: data.c,v 1.20 2003/03/04 12:31:38 cheusov Exp $
  * 
  */
 
 #include "dictzip.h"
+
 #include <sys/stat.h>
+#ifdef HAVE_MMAP
 #include <sys/mman.h>
+#endif
 #include <ctype.h>
 #include <fcntl.h>
+#include <assert.h>
+#ifdef HAVE_MMAP
+#include <sys/mman.h>
+#endif
 
 #define USE_CACHE 1
 
+#ifdef HAVE_MMAP
+int mmap_mode = 1; /* dictd uses mmap() function (the default) */
+#else
+int mmap_mode = 0;
+#endif
+
 int dict_data_filter( char *buffer, int *len, int maxLength,
 		      const char *filter )
 {
@@ -94,6 +107,7 @@ static int dict_read_header( const char *filename,
 	 }
       }
       header->crc = crc;
+      fclose( str );
       return 0;
    }
    header->type = DICT_GZIP;
@@ -206,10 +220,15 @@ static int dict_read_header( const char *filename,
 
 dictData *dict_data_open( const char *filename, int computeCRC )
 {
-   dictData    *h = xmalloc( sizeof( struct dictData ) );
+   dictData    *h = NULL;
    struct stat sb;
    int         j;
 
+   if (!filename)
+      return NULL;
+
+   h = xmalloc( sizeof( struct dictData ) );
+
    memset( h, 0, sizeof( struct dictData ) );
    h->initialized = 0;
 
@@ -226,16 +245,32 @@ dictData *dict_data_open( const char *filename, int computeCRC )
    
    if ((h->fd = open( filename, O_RDONLY )) < 0)
       err_fatal_errno( __FUNCTION__,
-		       "Cannot open index file \"%s\"\n", filename );
+		       "Cannot open data file \"%s\"\n", filename );
    if (fstat( h->fd, &sb ))
       err_fatal_errno( __FUNCTION__,
-		       "Cannot stat index file \"%s\"\n", filename );
+		       "Cannot stat data file \"%s\"\n", filename );
    h->size = sb.st_size;
 
-   h->start = mmap( NULL, h->size, PROT_READ, MAP_SHARED, h->fd, 0 );
-   if ((void *)h->start == (void *)(-1))
-      err_fatal_errno( __FUNCTION__,
-		       "Cannot mmap index file \"%s\"\b", filename );
+   if (mmap_mode){
+#ifdef HAVE_MMAP
+      h->start = mmap( NULL, h->size, PROT_READ, MAP_SHARED, h->fd, 0 );
+      if ((void *)h->start == (void *)(-1))
+	 err_fatal_errno(
+	    __FUNCTION__,
+	    "Cannot mmap data file \"%s\"\n", filename );
+#else
+      err_fatal (__FUNCTION__, "This should not happen");
+#endif
+   }else{
+      h->start = xmalloc (h->size);
+      if (-1 == read (h->fd, (char *) h->start, h->size))
+	 err_fatal_errno (
+	    __FUNCTION__,
+	    "Cannot read data file \"%s\"\n", filename );
+
+      close (h -> fd);
+      h -> fd = 0;
+   }
 
    h->end = h->start + h->size;
 
@@ -251,11 +286,25 @@ dictData *dict_data_open( const char *filename, int computeCRC )
 
 void dict_data_close( dictData *header )
 {
+   int i;
+
+   if (!header)
+      return;
+
    if (header->fd >= 0) {
-      munmap( (void *)header->start, header->size );
-      close( header->fd );
-      header->fd = 0;
-      header->start = header->end = NULL;
+      if (mmap_mode){
+#ifdef HAVE_MMAP
+	 munmap( (void *)header->start, header->size );
+	 close( header->fd );
+	 header->fd = 0;
+	 header->start = header->end = NULL;
+#else
+	 err_fatal (__FUNCTION__, "This should not happen");
+#endif
+      }else{
+	 if (header -> start)
+	    xfree ((char *) header -> start);
+      }
    }
 
    if (header->chunks)       xfree( header->chunks );
@@ -268,12 +317,49 @@ void dict_data_close( dictData *header )
 		       header->zStream.msg );
    }
 
+   for (i = 0; i < DICT_CACHE_SIZE; ++i){
+      if (header -> cache [i].inBuffer)
+	 xfree (header -> cache [i].inBuffer);
+   }
+
    memset( header, 0, sizeof( struct dictData ) );
    xfree( header );
 }
 
-char *dict_data_read( dictData *h, unsigned long start, unsigned long size,
-		      const char *preFilter, const char *postFilter )
+char *dict_data_obtain (const dictDatabase *db, const dictWord *dw)
+{
+   char *word_copy;
+   int len;
+
+   if (!dw)
+      return NULL;
+
+   if (dw -> def){
+      if (-1 == dw -> def_size){
+	 len = strlen (dw -> def);
+      }else{
+	 len = dw -> def_size;
+      }
+
+      word_copy = xmalloc (2 + len);
+      memcpy (word_copy, dw -> def, len);
+      word_copy [len + 0] = '\n';
+      word_copy [len + 1] = 0;
+
+      return word_copy;
+   }else{
+      assert (db);
+      assert (db -> data);
+
+      return dict_data_read_ (
+	 db -> data, dw -> start, dw -> end,
+	 db->prefilter, db->postfilter);
+   }
+}
+
+char *dict_data_read_ (
+   dictData *h, unsigned long start, unsigned long size,
+   const char *preFilter, const char *postFilter )
 {
    char          *buffer, *pt;
    unsigned long end;
@@ -293,7 +379,8 @@ char *dict_data_read( dictData *h, unsigned long start, unsigned long size,
    PRINTF(DBG_UNZIP,
 	  ("dict_data_read( %p, %lu, %lu, %s, %s )\n",
 	   h, start, size, preFilter, postFilter ));
-   
+
+   assert( h != NULL);
    switch (h->type) {
    case DICT_GZIP:
       err_fatal( __FUNCTION__,
