@@ -1,7 +1,7 @@
 /* dictd.h -- Header file for dict program
  * Created: Fri Dec  2 20:01:18 1994 by faith@dict
- * Revised: Wed Nov  8 05:57:21 2000 by faith@dict.org
- * Copyright 1994-2000 Rickard E. Faith (faith@dict.org)
+ * Revised: Mon Apr 22 15:47:26 2002 by faith@dict.org
+ * Copyright 1994-2000, 2002 Rickard E. Faith (faith@dict.org)
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -26,10 +26,16 @@
 #include "maa.h"
 #include "zlib.h"
 #include "codes.h"
+#include "dictdplugin.h"
 
 #include "net.h"
 #include <arpa/inet.h>
 #include <errno.h>
+
+#ifdef USE_PLUGIN
+#include <ltdl.h>
+#endif
+
 #include <netdb.h>
 #include <netinet/in.h>
 #include <signal.h>
@@ -49,10 +55,18 @@
 #define DICT_DAEMON_LIMIT        100 /* maximum simultaneous daemons */
 #define DICT_PERSISTENT_PRESTART 3 /* not implemented */
 #define DICT_PERSISTENT_LIMIT    5 /* not implemented */
-#define DICT_SHORT_ENTRY_NAME    "00-database-short"
-#define DICT_LONG_ENTRY_NAME     "00-database-long"
-#define DICT_INFO_ENTRY_NAME     "00-database-info"
-#define DICT_DEFAULT_STRATEGY    DICT_LEVENSHTEIN
+
+#define DICT_ENTRY_PREFIX        "00-database"
+#define DICT_ENTRY_PREFIX_LEN    sizeof(DICT_ENTRY_PREFIX)-1
+#define DICT_SHORT_ENTRY_NAME    DICT_ENTRY_PREFIX"-short"
+#define DICT_LONG_ENTRY_NAME     DICT_ENTRY_PREFIX"-long"
+#define DICT_INFO_ENTRY_NAME     DICT_ENTRY_PREFIX"-info"
+
+#define DICT_FLAG_UTF8           DICT_ENTRY_PREFIX"-utf8"
+#define DICT_FLAG_8BIT           DICT_ENTRY_PREFIX"-8bit"
+#define DICT_FLAG_ALLCHARS       DICT_ENTRY_PREFIX"-allchars"
+#define DICT_FLAG_VIRTUAL        DICT_ENTRY_PREFIX"-virtual"
+
 
 				/* End of configurable things */
 
@@ -82,8 +96,8 @@
 #define LOG_CLIENT      (0<<30|1<< 6) /* Log client                         */
 #define LOG_HOST        (0<<30|1<< 7) /* Log remote host name               */
 #define LOG_TIMESTAMP   (0<<30|1<< 8) /* Log with timestamps                */
-#define LOG_SYSLOG      (0<<30|1<< 9) /* Log using syslog(3)                */
-#define LOG_MIN         (0<<30|1<<10) /* Log a few minimal things           */
+#define LOG_MIN         (0<<30|1<< 9) /* Log a few minimal things           */
+#define LOG_AUTH        (0<<30|1<<10) /* Log authentication denials         */
 
 #define DICT_LOG_TERM    0
 #define DICT_LOG_DEFINE  1
@@ -92,6 +106,8 @@
 #define DICT_LOG_CLIENT  4
 #define DICT_LOG_TRACE   5
 #define DICT_LOG_COMMAND 6
+#define DICT_LOG_AUTH    7
+#define DICT_LOG_CONNECT 8
 
 #define DICT_UNKNOWN    0
 #define DICT_TEXT       1
@@ -137,29 +153,73 @@ typedef struct dictData {
    dictCache     cache[DICT_CACHE_SIZE];
 } dictData;
 
+typedef struct dictPlugin {
+#ifdef USE_PLUGIN
+   lt_dlhandle handle;
+#endif
+
+   void *      data;
+   /*   int         status;*/
+
+   dictdb_open_type   dictdb_open;
+   dictdb_set_type    dictdb_set;
+   dictdb_search_type dictdb_search;
+   dictdb_free_type   dictdb_free;
+   dictdb_error_type  dictdb_error;
+   dictdb_close_type  dictdb_close;
+} dictPlugin;
+
 typedef struct dictIndex {
    int           fd;		 /* file descriptor */
    const char    *start;	 /* start of mmap'd area */
    const char    *end;		 /* end of mmap'd area */
    unsigned long size;		 /* size of mmap */
-   const char    *optStart[257]; /* Optimized starting points */
+   const char    *optStart[UCHAR_MAX+2]; /* Optimized starting points */
    unsigned long headwords;	 /* computed number of headwords */
+
+   int    flag_utf8;         /* not zero if it has 00-database-utf8 entry*/
+   int    flag_8bit;         /* not zero if it has 00-database-8bit entry*/
+   int    flag_allchars;     /* not zero if it has 00-database-allchars entry*/
+
+   const int     *isspacealnum;
 } dictIndex;
 
 typedef struct dictDatabase {
    const char *databaseName;
    const char *databaseShort;
-   const char *databaseInfoPointer;
+   const char *databaseInfo;
    const char *dataFilename;
    const char *indexFilename;
+   const char *indexsuffixFilename;
+   const char *indexwordFilename;
    const char *filter;
    const char *prefilter;
    const char *postfilter;
    lst_List   acl;
    int        available;	/* if user has authenticated for database */
-   
+
    dictData   *data;
    dictIndex  *index;
+   dictIndex  *index_suffix;
+   dictIndex  *index_word;
+
+   lst_List   *virtual_db_list;
+
+   int invisible;    /* non-zero for invisible databases */
+
+   int exit_db;      /* non-zero for dictionary_exit entry */
+   int virtual_db;   /* non-zero for virtual databases */
+   int plugin_db;    /* non-zero for plugin entry */
+   int normal_db;    /* non-zero for normal database */
+
+   /* database_virtual members */
+   const char *database_list;  /* comma-separated list of database names */
+
+   /* database_plugin members */
+   const char *pluginFilename;
+   const char *plugin_data;    /* data for initializing plugin */
+   dictPlugin *plugin;
+
 } dictDatabase;
 
 #define DICT_DENY     0
@@ -167,6 +227,8 @@ typedef struct dictDatabase {
 #define DICT_AUTHONLY 2
 #define DICT_USER     3
 #define DICT_GROUP    4		/* Not implemented */
+#define DICT_MATCH    5         /* For IP matching routines */
+#define DICT_NOMATCH  6         /* For IP matching routines */
 
 typedef struct dictAccess {
    int        type;		/* deny, allow, accessonly, user, group */
@@ -180,21 +242,18 @@ typedef struct dictConfig {
    const char    *site;
 } dictConfig;
 
-#define DICT_EXACT        1	/* Exact */
-#define DICT_PREFIX       2	/* Prefix */
-#define DICT_SUBSTRING    3	/* Substring */
-#define DICT_SUFFIX       4	/* Suffix */
-#define DICT_RE           5	/* POSIX 1003.2 (modern) regular expressions */
-#define DICT_REGEXP       6	/* old (basic) regular expresions */
-#define DICT_SOUNDEX      7	/* Soundex */
-#define DICT_LEVENSHTEIN  8	/* Levenshtein */
-
-
 typedef struct dictWord {
-   const char    *word;
+   const dictDatabase  *database;
+   const dictDatabase  *database_visible;
+
+   char    *word;
+
    unsigned long start;
    unsigned long end;
-   dictDatabase  *database;
+
+/* Used by plugins */
+   const char    *def;
+   int            def_size;
 } dictWord;
 
 typedef struct dictToken {
@@ -206,25 +265,52 @@ typedef struct dictToken {
 extern dictData *dict_data_open( const char *filename, int computeCRC );
 extern void     dict_data_close( dictData *data );
 extern void     dict_data_print_header( FILE *str, dictData *data );
-extern int      dict_data_zip( const char *inFilename, const char *outFilename,
-			       const char *preFilter, const char *postFilter );
-extern char     *dict_data_read( dictData *data,
-				 unsigned long start, unsigned long end,
-				 const char *preFilter,
-				 const char *postFilter );
-extern int      dict_data_filter( char *buffer, int *len, int maxLength,
-				  const char *filter );
+extern int      dict_data_zip(
+   const char *inFilename, const char *outFilename,
+   const char *preFilter, const char *postFilter );
 
+extern char *dict_data_obtain (
+   const dictDatabase *db, const dictWord *dw);
+extern char *dict_data_read_ (
+   dictData *data,
+   unsigned long start, unsigned long end,
+   const char *preFilter,
+   const char *postFilter );
 
+extern int   dict_data_filter(
+   char *buffer, int *len, int maxLength,
+   const char *filter );
+
+/**/
 extern const char *dict_index_search( const char *word, dictIndex *idx );
-extern int        dict_search_database( lst_List l,
-					const char *word,
-					dictDatabase *database, int strategy );
-extern dictIndex  *dict_index_open( const char *filename );
+extern int         dict_search (
+   lst_List l,
+   const char *word,
+   const dictDatabase *database, int strategy,
+   int *extra_result,                  /* may be NULL */
+   const dictPluginData **extra_data,  /* may be NULL */
+   int *extra_data_size);              /* may be NULL */
+extern int dict_search_databases (
+   lst_List *l,
+   lst_Position db_pos,
+   const char *databaseName, const char *word, int strategy,
+   int *db_found);
+
+extern dictIndex  *dict_index_open(
+   const char *filename,
+   int init_flags, int flag_utf8, int flag_allchars );
 extern void       dict_index_close( dictIndex *i );
+
 extern void       dict_dump_list( lst_List list );
 extern void       dict_destroy_list( lst_List list );
 
+extern int        dict_destroy_datum( const void *datum );
+
+#ifdef USE_PLUGIN
+extern int        dict_plugin_open (dictDatabase *db);
+extern void       dict_plugin_close (dictDatabase *db);
+#endif
+
 /* dictd.c */
 
 extern void       dict_initsetproctitle( int argc, char **argv, char **envp );
@@ -237,12 +323,26 @@ extern dictConfig *DictConfig;  /* GLOBAL VARIABLE */
 extern int        _dict_comparisons; /* GLOBAL VARIABLE */
 extern int        _dict_forks;	/* GLOBAL VARIABLE */
 
+/*
+  If the filename doesn't start with / or .,
+  it is prepended with DICT_DIR
+*/
+extern const char *postprocess_dict_filename (const char *fn);
+/*
+  If the filename doesn't start with / or .,
+  it is prepended with PLUGIN_DIR
+*/
+extern const char *postprocess_plugin_filename (const char *fn);
+
 /* daemon.c */
 
 extern int  dict_daemon( int s, struct sockaddr_in *csin, char ***argv0,
 			 int delay, int error );
 extern void daemon_terminate( int sig, const char *name );
 
+/* */
+extern int        utf8_mode;
+
 				/* dmalloc must be last */
 #ifdef DMALLOC_FUNC_CHECK
 # include "dmalloc.h"
