@@ -1,7 +1,7 @@
 /* dictzip.c -- 
  * Created: Tue Jul 16 12:45:41 1996 by faith@dict.org
- * Revised: Fri Dec 22 06:04:56 2000 by faith@dict.org
- * Copyright 1996, 1997, 1998, 2000 Rickard E. Faith (faith@dict.org)
+ * Revised: Mon Apr 22 11:40:17 2002 by faith@dict.org
+ * Copyright 1996-1998, 2000, 2002 Rickard E. Faith (faith@dict.org)
  * 
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -17,13 +17,21 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: dictzip.c,v 1.16 2000/12/22 14:15:25 faith Exp $
+ * $Id: dictzip.c,v 1.21 2003/01/19 17:26:46 cheusov Exp $
  * 
  */
 
 #include "dictzip.h"
 #include <sys/stat.h>
 
+#ifndef HAVE_SNPRINTF
+extern int snprintf(char *str, size_t size, const char *format, ...);
+#endif
+
+#ifndef HAVE_VSNPRINTF
+extern int vsnprintf(char *str, size_t size, const char *format, va_list ap);
+#endif
+
 void dict_data_print_header( FILE *str, dictData *header )
 {
    char        *date, *year;
@@ -303,7 +311,7 @@ static const char *id_string( const char *id )
    static char buffer[BUFFERSIZE];
    char        *pt;
 
-   sprintf( buffer, "%s", DICT_VERSION );
+   snprintf( buffer, BUFFERSIZE, "%s", DICT_VERSION );
    pt = buffer + strlen( buffer );
 
    return buffer;
@@ -311,11 +319,11 @@ static const char *id_string( const char *id )
 
 static void banner( void )
 {
-   const char *id = "$Id: dictzip.c,v 1.16 2000/12/22 14:15:25 faith Exp $";
+   const char *id = "$Id: dictzip.c,v 1.21 2003/01/19 17:26:46 cheusov Exp $";
    
    fprintf( stderr, "%s %s\n", err_program_name(), id_string( id ) );
    fprintf( stderr,
-	    "Copyright 1996,1997 Rickard E. Faith (faith@cs.unc.edu)\n" );
+	    "Copyright 1996-2002 Rickard E. Faith (faith@dict.org)\n" );
 }
 
 static void license( void )
@@ -373,6 +381,7 @@ int main( int argc, char **argv )
 {
    int           c;
    int           i;
+   int           j;
    int           decompressFlag = 0;
    int           forceFlag      = 0;
    int           keepFlag       = 0;
@@ -385,6 +394,8 @@ int main( int argc, char **argv )
    char          *post          = NULL;
    unsigned long start          = 0;
    unsigned long size           = 0;
+   unsigned long clSize         = 0; /* from command line */
+   unsigned long clStart        = 0; /* from comment line */
    dictData      *header;
    char          *pt;
    FILE          *str;
@@ -417,7 +428,8 @@ int main( int argc, char **argv )
    dbg_register( DBG_ZIP,     "zip" );
    dbg_register( DBG_UNZIP,   "unzip" );
 
-   if (!(pt = strrchr( argv[0], '/' ))) pt = argv[0];
+   if ((pt = strrchr( argv[0], '/' ))) ++pt;
+   else                                pt = argv[0];
    if (!strcmp(pt, "dictunzip")) ++decompressFlag;
    if (!strcmp(pt, "dictzcat")) {
       ++decompressFlag;
@@ -443,10 +455,10 @@ int main( int argc, char **argv )
       case 'v': dbg_set( "verbose" );                                  break;
       case 'V': banner(); exit( 1 );                                   break;
       case 'D': dbg_set( optarg );                                     break;
-      case 's': ++decompressFlag; start = strtoul( optarg, NULL, 10 ); break;
-      case 'e': ++decompressFlag; size  = strtoul( optarg, NULL, 10 ); break;
-      case 'S': ++decompressFlag; start = b64_decode( optarg );        break;
-      case 'E': ++decompressFlag; size  = b64_decode( optarg );        break;
+      case 's': ++decompressFlag; clStart = strtoul( optarg, NULL, 10 ); break;
+      case 'e': ++decompressFlag; clSize  = strtoul( optarg, NULL, 10 ); break;
+      case 'S': ++decompressFlag; clStart = b64_decode( optarg );        break;
+      case 'E': ++decompressFlag; clSize  = b64_decode( optarg );        break;
       case 'p': pre = optarg;                                          break;
       case 'P': post = optarg;                                         break;
       default:  
@@ -456,6 +468,8 @@ int main( int argc, char **argv )
    if (testFlag) ++listFlag;
 
    for (i = optind; i < argc; i++) {
+      size  = clSize  ? clSize  : 0;
+      start = clStart ? clStart : 0;
       if (listFlag) {
 	 header = dict_data_open( argv[i], 1 );
 	 dict_data_print_header( stdout, header );
@@ -466,22 +480,22 @@ int main( int argc, char **argv )
 	    if (!size) size = header->length;
 	    if (!start) {
 	       len = header->chunkLength;
-	       for (i = 0; i < size; i += len) {
-		  if (i + len >= size) len = size - i;
-		  buf = dict_data_read( header, i, len, pre, post );
+	       for (j = 0; j < size; j += len) {
+		  if (j + len >= size) len = size - j;
+		  buf = dict_data_read_ ( header, j, len, pre, post );
 		  fwrite( buf, len, 1, stdout );
 		  fflush( stdout );
 		  xfree( buf );
 	       }
 	    } else {
-	       buf = dict_data_read( header, start, size, pre, post );
+	       buf = dict_data_read_ ( header, start, size, pre, post );
 	       fwrite( buf, size, 1, stdout );
 	       fflush( stdout );
 	       xfree( buf );
 	    }
 	    dict_data_close( header );
 	 } else {
-	    strcpy( filename, argv[i] );
+	    strncpy( filename, argv[i], BUFFERSIZE-1 );
 	    if ((pt = strrchr( filename, '.' ))) *pt = '\0';
 	    else
 	       err_fatal( __FUNCTION__, "Cannot truncate filename\n" );
@@ -493,9 +507,9 @@ int main( int argc, char **argv )
 	    header = dict_data_open( argv[i], 0 );
 	    if (!size) size = header->length;
 	    len = header->chunkLength;
-	    for (i = 0; i < size; i += len) {
-	       if (i + len >= size) len = size - i;
-	       buf = dict_data_read( header, i, len, pre, post );
+	    for (j = 0; j < size; j += len) {
+	       if (j + len >= size) len = size - j;
+	       buf = dict_data_read_ ( header, j, len, pre, post );
 	       fwrite( buf, len, 1, str );
 	       fflush( str );
 	       xfree( buf );
@@ -505,7 +519,7 @@ int main( int argc, char **argv )
 	       err_fatal_errno( __FUNCTION__, "Cannot unlink %s\n", argv[i] );
 	 }
       } else {
-	 sprintf( buffer, "%s.dz", argv[i] );
+	 snprintf( buffer,BUFFERSIZE-1, "%s.dz", argv[i] );
 	 if (!dict_data_zip( argv[i], buffer, pre, post )) {
 	    if (!keepFlag && unlink( argv[i] ))
 		err_fatal_errno( __FUNCTION__, "Cannot unlink %s\n", argv[i] );
